/*

	Simple program for Quantum Circuit synthesis using GA
	using minimization methods to reduce circuits length and cost
	-- possible modes
		- simple fitness function (only error)
		- complex fitness functin (scaled weighted sum of error and cost)
		- pareto replication
		- replicator USS (universal stochastic sampling)
		- roulette wheel

  -- coments to the names of programs
  -- example GS2simple2Mut
		- GS generic name of the program
		- 2 error is evaluated on bassis of both somponents of the matrix real and imaginary parts
		- simple/scaled/pareto type of fitness
		- 2 Mut - mutation operator is applied for each element of the sting representation

  -- all commands ar given through file



*/

#include "stdafx.h"
#include "GSE.h"
using namespace std;

typedef complex<double> dComp;


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


// temporary structures used only in incialization
struct tempT
{
       int numberGates;					// number of gates in this catagory
       char nameGate[MAXNUMOFGATES];	// for a particular number of input and output, the list of the gates can be used
};

// structure holding the quantum gates
struct qGate
{
       char representation, parentA, parentB;		// using a character to represent it,
       int numIO;									// number of input wires
	   int Cost;									// cost of this gate
       dComp gateMatrix1[MAXGATEDEM][MAXGATEDEM];	// matrix representing this qGate
};
// structure representing an individual in the GA population
struct Individual
{
       int ioNumber;		// number of wires of this individual
       int segmentNumber;	// number of segments at the creation, used only for intialization
       string my_string;	// circuit representation as string
       float fitness;		// fitness value
	   float Cost;			// Cost of the circuit
	   float Error;			// Error of the circuit
	   int Groups;			// for share fitness calcualtion
	   int Rank;
};


// holds GA functions
class GA
{
public:
	GA(){}
  ~GA(){};
  dComp I;								//complex variable
	// both here bellow fields will 
	// be in the future replaced by
	// array of pointers
  Individual *population[MAXNUMBER];		// population (array of object Individual)
  Individual *new_population[MAXNUMBER];	// population of individuals used for replication ... non optimal to be improved
  qGate *gateArray[MAXNUMOFGATES];			// allagtes are stored in here
  int numofgates;							// number of gates to be use < MAXNUMOFGATES
  Individual bestIndv;						// best Individual from the run
  ifstream in_stream;						// input file stream
  ofstream out_stream;						// output file stream

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
	
  // parameters for the GA
	// size of the population < MAXNUMOFGATES
  int populationNumber;
	// probabilities of mutation and crossover
  double proba_Mutation;
  double proba_Crossover;
	// scaling for the complex fitness function
	// alpha + beta = 1
  float alpha;
  float beta;
	//scaling fo the pareto replication
	//default alpha1 = 1, beta1 = 1
  float alpha1;
  float beta1;
	// estimated minimum of cost for a given circuit
  int divider;
	// type of GA Baldwinian or normal
	//0 - nomral, 1 - baldwinian
  int Ga;
	//setting different types of mutation
	//0 - normal, 1 - bitwise
  int mutation;
	//setting different types of crossover
	//0 - 1point, 1 - 2point
  int crossover;
	//different types of selection
	//0 - RW, 1- SUS 
  int replicator;
	// type of fitness function
	//0 - simple0, 1 - simple1, 2 - complex0, 3 - complex1
  int fitness;
	//method of fitness application
	//0 - normal, 1 - groupped shared
  bool grouped;
	//indicator of pareto rplication
	//0 - no, 1 - yes
  bool pareto;
	//Threshold for replication
	//0 - no threshold otherwise must be smaller than 1
  double threshold;
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

  
  
	//methods
  void initializePop();					// initialize the population
  void apply1Crossover(int, int);		// With a 1 point crossover probability recombine 2 parents into 2 new child
  void apply2Crossover(int, int);		// With a 2 point crossover probability recombine 2 parents into 2 new child
  void applyMutation(int);				// With a mutation probability mutate individual either normal or bitwise
  void applyReplication();				// Replication of teh selected individuals either SUS or RW
  void applyParetoReplication();		// Replication for pareto optimal GA
  void makeFitness(Individual*, qGate*, double, double);	// according to the input it calculates different fitness values based on the error and the cost 

  void setTerminatingCondition(int);	// somemethod setting final criteria
  bool terminatingCondition();			// some method returning if the final conditions were reached
  void individualMatrix(int, bool);		// calculation of the individual matrix representation for each individual
  void groupMatrix(int, bool);			// calculating matrix and also sharing fitness [EXPERIMENTAL]
  void clearAll(int);					// clear everything in one individual
  int minimizeCirc(Individual*);		// method minimizing circuit ... not used always
  void repairCircuit(int);				// method repairing circuit after mutation
  void closeStream();					// method close streams to files
  void output(int);						// writes down the result of the run
  void outputBest(int);					// writes down the best individual from one generation or the allover best
  Individual setIndv(Individual);		// copy one Individual and output it
  void setIndv(Individual*, Individual*); //copy two indidual by their addresses

private:

  qGate nullGate;						// nullGate - wire
  qGate finalGate;						// gate/circuit we are looking for
  tempT stringArray[MAXINDVWIRENUM];	// sorted by input wires in asending order
  int resultnum;						// the number of wire for input and output of the result gate
  int generationCondition;				// termination criterian
  bool condition;						// some boolean condition
  float best;

  qGate tensorProduct(qGate*, qGate*);	// tensor product of gate A and gate B
  qGate matrixProduct(qGate*, qGate*);	// matrix product of gate A and gate B
  int getGate(char);					// retunrs teh char-representation of the gate at teh int index from gateArray
  void reduceSegNumber(int);			// reduce teh total length of the circuit
  void initializeGates();				// reads teh gates from a file and set them in an array
  void makeIndvString(int);				// creates string representation for each individual
  void setStringArray();				// set an array used only for intialization
  bool getFitness();					// looks for the best individual with fitness 1

  //for minimizer of quantum circuits
  qGate copyGate(qGate*);						// makes a copy oif a quantum gate
  int compareGate(qGate*);						// compares one quantum gate to all available
  bool compareGate(qGate, qGate);				// compare two gates
  void removeRedund(Individual*);				// remove redundantsegments in the circuit
  basic_string<char> synthSegment(int, int);	// creates a new segment of gates in string representation
  void mergeSegs(Individual*, int, int);		// merge two equivalent or equal segments

};



/////////////////////////////////////////////////////////////////////////////
// The one and only application object

CWinApp theApp;


int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
	int nRetCode = 0;

	// initialize MFC and print and error on failure
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
	{
		// TODO: change error code to suit your needs
		cerr << _T("Fatal Error: MFC initialization failed") << endl;
		nRetCode = 1;
	}
	else
	{
		// TODO: code your application's behavior here.
		CString strHello;
		strHello.LoadString(IDS_HELLO);
		cout << (LPCTSTR)strHello << endl;


		
		srand( time(NULL) );
		int generationCondition = 0;

		//ignition
		GA g;
		g.initializePop();

		for (int a = 0;a<g.populationNumber;a++)
		{
			if (!g.grouped)
				g.individualMatrix(a,false);
			else g.groupMatrix(a, false);
		}
		
		cout<<"Initialization completely done"<<endl;
		g.setIndv(&(g.bestIndv), g.population[0]);
		int counter = 0;

		//life cycle
		while (!g.terminatingCondition())
		{
			cout<<" current generation is "<<generationCondition<<endl;
			for (a = 0;a<g.populationNumber;a++)
				if (!g.pareto)
					g.applyReplication();
				else g.applyParetoReplication();

			for (a = 0;a<g.populationNumber;a++)
			{
			
				g.applyMutation(a);
				//g.population[a].fitness = 0;
				g.repairCircuit(a);
				if (!g.grouped)
					g.individualMatrix(a,false);
				else g.groupMatrix(a, false);
			}

			if (counter == 100)
			{
				//cout<<"selecting the best individual"<<endl;
				g.outputBest(generationCondition);
				counter = 0;
			}

			generationCondition++;
			//cout<<"getting terminating condition"<<endl;
			g.setTerminatingCondition(generationCondition);
			counter++;
		} 

		//---------printing---------------
		g.output(generationCondition);
		g.closeStream();
	}
	return nRetCode;
}

/*
	
     calculation of individual fitness!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     and of individual matrix representation
	 setting the fitness for each individual also
	 because each individual stores only a representation 
	 string not the whole circuit

*/
void GA::individualMatrix(int i, bool display)
{    
	//temporary storage values
    float Error, Cost;
	int Time, a, b, j;
    //four free gates for operations
    qGate *temp0, *temp1, *temp2, *finalG;

	Error = 0;
	Cost = 0;
	Time = 0;
	Individual *I = population[i];

    if (I->ioNumber <= 1) 
    {
		j = 0;
		
		//setting the first gate
        for (a = 0;a<= numofgates;a++)
			if (gateArray[a]->representation == I->my_string.at(j))
			{
				temp0 = gateArray[a]; 
				Cost+= gateArray[a]->Cost; 
				break;
			}
  		j++;
		
		while (j<=I->my_string.length()-1)
		{
			//setting the next gate
			for (a = 0;a<= numofgates;a++)
				if (gateArray[a]->representation == I->my_string.at(j))
				{
					temp2 = gateArray[a]; 
					Cost+= gateArray[a]->Cost; 
					break;
				}

			//making matrix product on the location of the first gate
			//recursive calculation of the circuit
			temp0 = &matrixProduct(temp0, temp2);
			j++;
		}
        finalG = temp0;
	} else //for more than one wire circuit
	{
		int begin = 0;
		int end = I->my_string.find("p", begin+1);
		
		while(end > 0) 
		{
			if ((end - begin) > 1)
			{
				for (a = 0;a<= numofgates;a++)
					if (gateArray[a]->representation == I->my_string.at(begin+1)) 
					{
						temp0= gateArray[a];
						begin++; Cost+= gateArray[a]->Cost;
						break;
					}
					           
				for (b = begin+1;b<end;b++) 
				{
					for (a = 0;a<= numofgates;a++)
						if (gateArray[a]->representation == I->my_string.at(b)) 
						{
							temp1= gateArray[a];
							Cost+= gateArray[a]->Cost;
							break;
						}
					
					temp0 = &tensorProduct(temp0, temp1);
				}
				if (Time == 0) 
				{
					finalG = temp0;
					Time++;
				}
				else 
				{
					finalG = &matrixProduct(finalG, temp0); 
					Time++;
				}
			}
			begin = I->my_string.find("p", end);
			end = I->my_string.find("p", begin+1);
		}
	}
	finalG->numIO = I->ioNumber;
	//direct correspondence between number of wires in the goal and current individual
	if (finalG->numIO!= finalGate.numIO) 
	{
		I->fitness =  0;
	}
	else 
	{
		for (int a = 0; a<(int(pow(2,finalG->numIO)));a++)
		{
			for (int b = 0; b<(int(pow(2,finalG->numIO)));b++)
			{ 
				//simple comparison
				if (finalGate.gateMatrix1[a][b] != finalG->gateMatrix1[a][b])Error++;
			 }
		}



			makeFitness(I, finalG, Error, Cost);


		
	}
	if (display)
	{
		out_stream << " -------- " <<endl;
		for (int a = 0; a<(int(pow(2,finalG->numIO)));a++)
		{
			for (int b = 0; b<(int(pow(2,finalG->numIO)));b++)
				out_stream << finalG->gateMatrix1[a][b] <<" ";
			
			out_stream << endl;
		}
	}
}



void GA::groupMatrix(int i, bool display)
{    //temporary storage values
     float Error = 0;
	 int Time = 0;
	 float Cost = 0;
	 int proba1;
     //three free gates for operations
	 //finalG is the mtrix represenation for fitness evaluation
     qGate *temp0, *temp1, *temp2, *finalG;
	 Individual *I = population[i];

	 basic_string<char> tempStr;
	 
	 for (int b = 0;b<GROUPSIZE;b++)
	 {
		 tempStr = I->my_string;
		 do {
			 proba1 = (rand()%(populationNumber));
		 } while (I->ioNumber != population[proba1]->ioNumber);
		 
		 tempStr.append(population[proba1]->my_string);
		 
		 if (I->ioNumber <= 1) 
		 {
			 int j = 0;
			 for (int a = 0;a<= MAXNUMOFGATES;a++)
				 if (gateArray[a]->representation == tempStr.at(j))
				 {
					 temp0 = gateArray[a]; 
					 Cost+= gateArray[a]->Cost; 
					 break;
				 }
			 j++;


		   while (j<=tempStr.length()-1)
		   {
			   for ( a = 0;a<= MAXNUMOFGATES;a++)
				if (gateArray[a]->representation == tempStr.at(j))
				{
					temp2 = gateArray[a]; 
					Cost+= gateArray[a]->Cost; 
					break;
				}
				
				//making matrix product on the location of the first gate
				//recursive calculation of the circuit
				
				temp0 = &matrixProduct(temp0, temp2);
				j++;
		   }
		   finalG = temp0;
		 
		 } else //for more than one wire circuit
		 {
			 int begin = 0;
			 int end = tempStr.find("p", begin+1);
			 
			 while(end > 0) 
			 {
				 if ((end - begin) > 1)
				 {
					 for (int a = 0;a<= MAXNUMOFGATES;a++)
						 if (gateArray[a]->representation == tempStr.at(begin+1)) 
						 {
							 temp0= gateArray[a]; 
							 begin++; 
							 Cost+= gateArray[a]->Cost; 
							 break;
						 }
						 
						 for (int b = begin+1;b<end;b++) 
						 {
							 for (int a = 0;a<= MAXNUMOFGATES;a++)
								 if (gateArray[a]->representation == tempStr.at(b)) 
								 {
									 temp1= gateArray[a];
									 Cost+= gateArray[a]->Cost; 
									 break;
								 }
								 
							 temp0 = &tensorProduct(temp0, temp1);
						 }
						 if (Time == 0) 
						 {
							 finalG = temp0;
							 Time++;
						 }
						 else {
							 finalG = &matrixProduct(finalG, temp0); 
							 Time++;
						 }
				 }
				 begin = tempStr.find("p", end);
				 end = tempStr.find("p", begin+1);
			 }
		 }
		 
		 finalG->numIO = I->ioNumber;
		 
		 //direct correspondence between number of wires in the goal and current individual
		 if (finalG->numIO!= finalGate.numIO) 
			 I->fitness =  0;
		 else 
		 {
			 for (int a = 0; a<(int(pow(2,finalG->numIO)));a++)
				 for (int b = 0; b<(int(pow(2,finalG->numIO)));b++)
				 { 	
					 //simple comparison
					 if (finalGate.gateMatrix1[a][b] != finalG->gateMatrix1[a][b])
						 Error++;
				 }

			makeFitness(I, finalG, Error, Cost);
			I->my_string = tempStr;

			 
		 }
	}
	if (display)
	{
		out_stream << " -------- " <<endl;
		for (int a = 0; a<(int(pow(2,finalG->numIO)));a++)
		{
			for (int b = 0; b<(int(pow(2,finalG->numIO)));b++)
				out_stream << finalG->gateMatrix1[a][b] <<" ";
			out_stream << endl;
		}
	}

	if (i == populationNumber-1)
		for (int o = 0;o<populationNumber;o++)
		{
			population[o]->fitness/= population[o]->Groups;
		}
}


void GA::setStringArray()
{
	qGate *Q;
	int count[MAXNUMOFGATES];
	
	for (int j = 0; j < MAXGATEINPUT; j++)
	{
		count[j] = 0;
		for (int i = 0; i < numofgates; i++)
		{
			Q = gateArray[i];
			if (Q->numIO == (j+1))
			{
				stringArray[j].nameGate[count[j]] = Q->representation;
				count[j]++;
			}
			stringArray[j].numberGates = count[j];
		}
	}
}


void GA::makeIndvString(int i)
{
    int tempNumber,tempNumber2, tempGate, counter, temporyhold;
	char gate;
	Individual *I = population[i];
	counter = 0;
		
	for (int j = 0; j < I->segmentNumber; j++) // for each segment of this individual
	{
		if (I->ioNumber > MAXGATEINPUT) // if the input has more than MAXGATEINPUT (4)
			tempGate = MAXGATEINPUT;             // max gate to use
		else
			tempGate = I->ioNumber;
		tempNumber = I->ioNumber; // the number of inputs
		do
		{
			tempNumber2 = rand()%tempGate; // which gate to use
		}while (stringArray[tempNumber2].numberGates == 0);
		
		if (I->ioNumber > 1) // if not only one input
		{
			I->my_string += 'p';
			counter++;
		}
		
		while ((tempNumber-(tempNumber2+1)) > 0) // while more gate could be parallel
		{
			temporyhold = rand()%stringArray[tempNumber2].numberGates;
			gate = stringArray[tempNumber2].nameGate[temporyhold];
			
			I->my_string += gate;
			counter++;
			
			tempNumber = tempNumber - (tempNumber2+1);
			if (tempNumber > tempGate)              // use the smaller one of the two
			{
				temporyhold = rand()%stringArray[tempNumber2].numberGates;
				gate = stringArray[tempNumber2].nameGate[temporyhold];
				do
				{
					tempNumber2 = rand()%(tempGate);
				}while (stringArray[tempNumber2].numberGates == 0);
			}
			else
			{
				do
				{
					tempNumber2 = rand()%(tempNumber);
				}while (stringArray[tempNumber2].numberGates == 0);
			}
		}
		temporyhold = rand()%stringArray[tempNumber2].numberGates;
		gate = stringArray[tempNumber2].nameGate[temporyhold]; // when only can choose one input gate
		
		I->my_string += gate;
		counter++;
		
		if (I->ioNumber > 1)
		{
			I->my_string += 'p';
			counter++;
		}
	}
}




void GA::output(int generationCondition)
{
	qGate *qG;

     out_stream << "There are " << numofgates << " gates from the input file and synthesized.\n";
     for (int i = 0; i < numofgates; i++)
     {
		 qG = gateArray[i];

         out_stream << "The gate number " << i << " is: " << endl;
         out_stream << "And the gate representation is: " << gateArray[i]->representation << endl;
         out_stream << "Number of inputs is: " << qG->numIO << endl;
         out_stream << "Parent A is: " << qG->parentA << endl;
         out_stream << "Parent B is: " << qG->parentB << endl;

         for (int j = 0; j < (int(pow(2,qG->numIO))); j++)
         {
             for (int k = 0; k < (int(pow(2,qG->numIO))); k++)
             {
                 out_stream << qG->gateMatrix1[j][k] << " ";
             }
             out_stream << endl;
         }
     }
     out_stream << "Number of inputs is: " << finalGate.numIO << endl;
     out_stream << "The optimal gate is: \n";
     for (int m = 0; m < (int(pow(2,resultnum))); m++)
     {
         for (int n = 0; n < (int(pow(2,resultnum))); n++)
         {
             out_stream << finalGate.gateMatrix1[n][m] << " ";
         }
         out_stream << endl;
     }


      out_stream << "---------------------------" << endl;
      out_stream << " End of run at " <<generationCondition<< endl;
      out_stream << "---------------------------" << endl;
      out_stream << "----- Best Individual -----" << endl;
	  out_stream << bestIndv.ioNumber<<" number of inputs "<<endl;

	  setIndv(population[0], &bestIndv);
	  individualMatrix(0, true);

	  out_stream<<endl;
	  out_stream<< bestIndv.my_string<<" representation "<<endl;
	  out_stream<< bestIndv.fitness<<" fitness "<<endl;
}


void GA::outputBest(int generationCondition)
{
	setIndv(population[0], &bestIndv);
	individualMatrix(0, true);
	
	out_stream << "---------------------------" << endl;
	out_stream << " End of run at " <<generationCondition<< endl;
	out_stream << "---------------------------" << endl;
	out_stream << "----- Best Individual -----" << endl;
	out_stream << bestIndv.ioNumber<<" number of inputs "<<endl;
	out_stream<<endl;
	out_stream<< bestIndv.my_string<<" representation "<<endl;
	out_stream<< bestIndv.fitness<<" fitness "<<endl;
	out_stream<< bestIndv.Error<<" error "<<endl;
	out_stream<< bestIndv.Cost<<" Cost "<<endl;
}

// Restrictions:
// 1. no fan out - ex, 5 inputs has to have 5 outputs
// 2. maximum 5 I/O (geometric increase of complexity with more I/Os)
void GA::initializePop()
{
     char inFile[100];
     char outFile[100];
     condition= false;

     // Open input and output files.
     cout << "Please input the source file" << endl;
     cin >> inFile;


     in_stream.open(inFile);
     if (in_stream.fail())
     {
        cout << "Input file opening failed.\n";
        exit(1);
     }

     cout << "\nPlease input the output file" << endl;
     cin >> outFile;

     out_stream.open(outFile);
     if (out_stream.fail())
     {
        cout << "Output file opening failed.\n";
        exit (1);
     }

     initializeGates();
     setStringArray();

     for (int i = 0; i < populationNumber; i++)
     {
		 population[i] = new Individual;
		 new_population[i] = new Individual;	
		 population[i]->ioNumber = finalGate.numIO;
         population[i]->segmentNumber = rand()%MAXSEGNUM+1; 
		 makeIndvString(i);
     }
    out_stream << "\nintialization done" << endl;

}



int GA::getGate(char b)
{
	for (int a = 0;a<numofgates;a++)
		if (gateArray[a]->representation == b) return a;
  
}


// Restrictions:
// 1. mutate one gate to another with same number of I/O
// 2. mutate 2 or more parallel connections between them
// 3. nutate any gate with n I/O to any gates or gate with same n I/O
void GA::applyMutation(int indv)
{

	qGate temp, temp1, old;		
    char replaceGate, gate;
	int proba2, length;
	float proba;
	Individual *I = population[indv];

	///////////////////////
	//setting mutation for either bitwise either normal
	if (mutation != 0)
		length = I->my_string.length();
	else 
		length = 1;
	///////////////////////
	///////////////////////

	for (int A = 0; A<length;A++)
	{
		proba  = 1-(double(1.0*rand()/((double)RAND_MAX+1.0)));
		
		if (proba <= proba_Mutation)
		{
			proba2 =  rand()%(*I).my_string.length();
			gate =  I->my_string.at(proba2);
			proba  = 1-(double(1.0*rand()/((double)RAND_MAX+1.0)));

			if (proba < 0.33)
			{
				if (gate == 'p' || proba2>(I->my_string.length() -1))
				{   
					do {
						proba2 = rand()%I->my_string.length();
						gate =  I->my_string.at(proba2);
					}while(gate == 'p' || proba2>(I->my_string.length() -1));
				}
				int o = getGate(gate);
				old = *gateArray[o];
				do {
					temp = *gateArray[rand()%numofgates];
				}while(temp.numIO > I->ioNumber || temp.numIO <= 0); 
				
				if (old.numIO == temp.numIO) 
				{
					I->my_string.erase(proba2, 1);
					I->my_string.insert(proba2, 1, temp.representation);
				} 
				else
					//we need to find more gates to fill the hole (forced version)
					if (old.numIO > temp.numIO) 
					{
						int diff = old.numIO - temp.numIO;
						do {
							temp1 = *gateArray[rand()%numofgates];
						}while(temp1.numIO != diff);
						
						I->my_string.erase(proba2, 1);
						I->my_string.insert(proba2, 1, temp.representation);
						I->my_string.insert(proba2+1, 1, temp1.representation);
					} 
					else
						//we need to create a new parallel box 
						if (old.numIO < temp.numIO) 
						{
							int A = 1;
							int tempinput = 0;
							int tempinputb = 0;
							
							replaceGate = (*gateArray[0]).representation;
							//first i replace the selected gate
							I->my_string.erase(proba2, 1);
							I->my_string.insert(proba2, 1, temp.representation);
							
							//mapping the neighborhood
							//setting how many missing connections i will hve to replace after me
							while(I->my_string.at(proba2+A) != 'p') 
							{
								tempinput += (*gateArray[getGate(I->my_string.at(proba2+A))]).numIO;
								A++; 
							}
							
							A = 1;
							
							//setting how many missing connections i will hve to replace behind me
							while(I->my_string.at(proba2-A) != 'p') 
							{
								tempinputb += (*gateArray[getGate(I->my_string.at(proba2-A))]).numIO;
								A++; 
							}
							
							//	inserting gates behind me if someone behind me
							if (tempinputb > 0 && (tempinputb+temp.numIO > I->ioNumber))
							{
								//inserting empty gates behind my position
								for (int a = 0;a<(I->ioNumber - tempinputb);a++) 
									I->my_string.insert(proba2, 1, replaceGate); 
								
								proba2 += (I->ioNumber - tempinputb);
								//insering a new cell individual
								for (a = 0;a<2;a++) 
									I->my_string.insert(proba2, "p");
								
								proba2 += 2;
								tempinputb = 0;
							
							} 
							else 
								tempinputb += temp.numIO;
							if ((tempinputb +tempinput) > I->ioNumber)
							{
								proba2++;
								
								//insereting after me
								for (int a = 0;a<(I->ioNumber - tempinputb);a++)
									I->my_string.insert(proba2+a, 1, replaceGate);
								
								proba2 += (I->ioNumber - tempinputb);
								//creating new cell
								for (a = 0;a<2;a++)
									I->my_string.insert(proba2+a, "p");
								
								proba2 += 2;
								//filling next cell with wires if necessarry
								for (a = 0;a<(I->ioNumber - tempinput);a++)
									I->my_string.insert(proba2+a, 1, replaceGate);
							} else 
								if ((tempinputb + tempinput) < I->ioNumber)
								{
									tempinput += temp.numIO;
									//fill emty space with wires
									for (int a = 0;a<(I->ioNumber - tempinput);a++)
										I->my_string.insert(proba2+a, 1, replaceGate);
								}
						} 
		} else
			if (proba >= 0.33 && proba < 0.66)
			{
				if (I->my_string.length() > (I->ioNumber*3))
				{
					if (I->ioNumber <= 1)
					{
						proba2 = (rand()%I->my_string.length())-1;
						if (proba2 < 0) proba2 = 0;
						I->my_string.erase(proba2, 1);
					} else
					{
						do {
							proba2 = (rand()%I->my_string.length())-1;
							if (proba2 < 0) proba2 = 0;
							gate =  I->my_string.at(proba2);
						}while(gate != 'p' || proba2>(I->my_string.length() -3));
						
						if (I->my_string.at(proba2+1) == 'p') proba2++;
						int proba3 = I->my_string.find("p", proba2+1);
						I->my_string.erase(proba2, (proba3-proba2+1));
					}
				}
			}
			else
			{
				basic_string<char> tempS;
				int V = 0;
				do {
					temp = *gateArray[rand()%numofgates];
					if (temp.numIO <= (I->ioNumber-V) && temp.numIO > 0)
					{
						V += temp.numIO;
						tempS.insert(tempS.length(), 1, temp.representation);
					}
				}while(V != I->ioNumber);  
				tempS.insert(0, "p");
				tempS.insert(tempS.length(), "p");
				
				I->my_string.insert(I->my_string.length(), tempS);
			}
		}
	}
}


void GA::repairCircuit(int a)
{
	int begin, end, counter;
	bool condition, erased;
	Individual *I = population[a];
	condition = false;
	begin = 0;
	
	//for mal formed circuits --- very malformed
	if (I->my_string.size()<3)
	{
		I->ioNumber = finalGate.numIO;
		I->segmentNumber = rand()%MAXSEGNUM+1; // number of segments in series, between 1 and MAXSEGNUM
		makeIndvString(a); // create an string for each individual in the population
		condition = true;
	}

	while (!condition)
	{
		counter = 0;
		erased = false;
		end = I->my_string.find("p", begin+1);
		
		if (end > I->my_string.size() || end < 0)
			break;


		if (I->my_string.at(0) != 'p' || I->my_string.at(I->my_string.size()-1) != 'p' )
		{
			 I->ioNumber = finalGate.numIO;
			 I->segmentNumber = rand()%MAXSEGNUM+1; // number of segments in series, between 1 and MAXSEGNUM
			 makeIndvString(a); // create an string for each individual in the population
			 condition = true;
			 break;
		}

		for (int n = begin;n<end;n++)
			if (I->my_string.at(n) != 'p') 
			{	
				counter += (*gateArray[getGate(I->my_string.at(n))]).numIO;
			}

		if (counter < I->ioNumber )
		{

			I->my_string.erase(begin, (end-begin+1));
			erased = true;
			

		}
		
		if (erased)
		{
			if (I->my_string.length()<1)
			{
				 I->ioNumber = finalGate.numIO;
				 I->segmentNumber = rand()%MAXSEGNUM+1; // number of segments in series, between 1 and MAXSEGNUM
				 makeIndvString(a); // create an string for each individual in the population
				 condition = true;
			}

		} else 
		{
			if (end < I->my_string.length()-2)
				begin = I->my_string.find("p",end+1);
			else condition = true;
		}
	}
}




// Restrictions:
// 1. replace gate with same number of I/O
// 2. replace one gate with n I/O with any combination of smaller gates with
//    same number of I/O
// 3. conserve strict correspondence of connections
//                                void operator=(const class bla & other) { mymember=other.mymember; }
void GA::apply1Crossover(int ind1, int ind2)
{  
     //	apply rules and make crossover
	int proba1, proba2;
	Individual *I = new_population[ind1];
	Individual *J = new_population[ind2];

	if (I->ioNumber == J->ioNumber)
	{
		do {
			proba1 = (rand()%(I->my_string.length()));
		}while(proba1 >= I->my_string.length());

		do {
			proba2 = (rand()%(J->my_string.length()));
		}while(proba2 >= J->my_string.length());

	    basic_string<char> temp1;
		basic_string<char> temp2;


		if (I->ioNumber > 1) 
		{ 
			if (proba1 == 0) temp1 = I->my_string.substr(proba1);
			else 
			{
				if ((I->my_string.at(proba1)) == 'p')
					if ((I->my_string.at(proba1-1)) == 'p') proba1-=2;
					else proba1--;
					
				while((I->my_string.at(proba1)) != 'p')
					proba1--;

				temp1 = I->my_string.substr(proba1);
			}
			
			if (proba2 == 0) temp2 = J->my_string.substr(proba2);
			else 
			{
				if ((J->my_string.at(proba2)) == 'p')
					if ((J->my_string.at(proba2-1)) == 'p') 
						proba2-=2;
					else proba2--;

				while((J->my_string.at(proba2)) != 'p')
					proba2--;
				temp2 = J->my_string.substr(proba2);
			}

			if (temp1.length() >=3 && temp2.length() >= 3)
			{
				I->my_string.replace(proba1, temp1.length(), temp2, 0, temp2.length());
				J->my_string.replace(proba2, temp2.length(), temp1, 0, temp1.length());
			}

	   } else
	   {

			temp1 = I->my_string.substr(proba1);
			temp2 = J->my_string.substr(proba2);

			if (temp1.length() > 1 && temp2.length() > 1)
			{
				I->my_string.replace(proba1, I->my_string.length(), temp2, 0, temp2.length());
				J->my_string.replace(proba2, J->my_string.length(), temp1, 0, temp1.length());
			}
	   }
	}
}

void GA::apply2Crossover(int ind1, int ind2)
{  
     //	apply rules and make crossover
	int proba[4], a;
	Individual *I = new_population[ind1];
	Individual *J = new_population[ind2];

	if (I->ioNumber == J->ioNumber)
	{
		do {
			proba[0] = (rand()%(I->my_string.length()));
		}while(proba[0] >= (I->my_string.length())/2);

		do {
			proba[1] = (rand()%(I->my_string.length()));
		}while(proba[1] >= I->my_string.length() && proba[1] <= proba[0]);

		do {
			proba[2] = (rand()%(J->my_string.length()));
		}while(proba[2] >= (J->my_string.length())/2);


		do {
			proba[3] = (rand()%(J->my_string.length()));
		}while(proba[3] >= J->my_string.length() && proba[3] <= proba[2]);

	    basic_string<char> temp1;
		basic_string<char> temp2;

		for (a = 0;a<4;a++)
		{
			if (a == 0)
			{
				while(true)
				{
					if (I->my_string.at(proba[a]) != 'p')
						proba[a]--;
					if (I->my_string.at(proba[a+1]) != 'p')
						proba[a+1]++;
					if (I->my_string.at(proba[a]) == 'p' && I->my_string.at(proba[a+1]) == 'p')
						break;
				}
			}
			else
				if (a == 2)
					while(true)
					{
						if (I->my_string.at(proba[a]) != 'p')
							proba[a]--;
						if (I->my_string.at(proba[a+1]) != 'p')
							proba[a+1]++;
						if (I->my_string.at(proba[a]) == 'p' && I->my_string.at(proba[a+1]) == 'p')
							break;
					}
		}

		if (I->ioNumber > 1) 
		{ 

			
			temp1 = I->my_string.substr(proba[0], proba[1]-proba[0]+1);
			temp2 = J->my_string.substr(proba[2], proba[3]-proba[2]+1);


			if (temp1.length() >=3 && temp2.length() >= 3)
			{
				out_stream<<I->my_string<<endl;
				out_stream<<J->my_string<<endl;

				I->my_string.replace(proba[0], temp1.length(), temp2, 0, temp2.length());
				J->my_string.replace(proba[2], temp2.length(), temp1, 0, temp1.length());

				out_stream<<I->my_string<<endl;
				out_stream<<J->my_string<<endl;

			}

	   } else
	   {

			temp1 = I->my_string.substr(proba[0], proba[1]);
			temp2 = J->my_string.substr(proba[2], proba[3]);

			if (temp1.length() > 1 && temp2.length() > 1)
			{
				I->my_string.replace(proba[0], I->my_string.length(), temp2, 0, temp2.length());
				J->my_string.replace(proba[2], J->my_string.length(), temp1, 0, temp1.length());
			}
	   }
	}
}


void GA::applyReplication()
{ 

    float fitnessSum = 0;
	float first, second, tempsum, proba;
	int indv1, indv2;

    for (int i = 0; i < populationNumber; i++)
		fitnessSum += population[i]->fitness;

    int counter = 0;
    bool condition = false;
    while (!condition)
	{
		if (replicator == 0)
		{
			//roulette wheel
			if (threshold > 0)
				do{
					indv1  = rand()%(populationNumber-1);
					indv2  = rand()%(populationNumber-1);
				} while(population[indv1]->fitness > threshold &&  population[indv2]->fitness > threshold);
			else
				{
					first  = (1-(double(1.0*rand()/((double)RAND_MAX+1.0))))*fitnessSum;
					second  = (1-(double(1.0*rand()/((double)RAND_MAX+1.0))))*fitnessSum;
				}
		}
		else
		{
			//stochastic uniersal sampling
			if (threshold > 0)
				do{
					indv1  = rand()%(populationNumber-1);
					indv2  = rand()%(populationNumber-1);
				} while(indv1 > threshold &&  indv2 > threshold);
			else
				{
					first  = ((1-(double(1.0*rand()/((double)RAND_MAX+1.0))))*fitnessSum)/2;
					second = fitnessSum - first;
				}
		}

		if (threshold == 0)
		{
			tempsum = 0;
			for (int c = 0;c<populationNumber;c++)
			{
				tempsum += population[c]->fitness;
				if (tempsum >= first) 
				{
					indv1 = c; 
					break;
				}
			}
			
			tempsum = 0;
			for (int d = 0;d<populationNumber;d++)
			{
				tempsum += population[d]->fitness;
				if (tempsum >= second) 
				{
					indv2 = d; 
					break;
				}
			}
		}
		
		if(counter == populationNumber-1)
		{
			setIndv(new_population[counter], population[indv1]);
			counter++;
		} else
		{
			setIndv(new_population[counter], population[indv1]);
			setIndv(new_population[counter+1], population[indv2]);

			proba  = 1-(double(1.0*rand()/((double)RAND_MAX+1.0)));
			if (proba < proba_Crossover && indv1 != indv2)
			{
				if (crossover == 1)
					apply1Crossover(counter, counter+1);
				else apply2Crossover(counter, counter+1);
			}
			
			counter += 2;
		}
		if (counter >= populationNumber-1) condition = true;
	}

    for (i = 0;i<populationNumber;i++)
	{
	   setIndv(population[i], new_population[i]);
	 
	}
}


void GA::applyParetoReplication()
{ 
	float proba;
	int indv1, indv2, tempsum, fitnessSum, first, second;
	int counter = 0;
    bool condition = false;

	for (int i = 0; i < populationNumber; i++) population[i]->Rank = -1;
	int temp_rank;
	
	for (i = 0; i < populationNumber; i++)
    {
		temp_rank = 0;
		for (int j = 0; j < populationNumber; j++)
		{
			if (population[i]->Rank <= -1)
			{
				if ((alpha1*population[i]->Error) <= population[j]->Error && (beta1*population[i]->Cost) <= population[j]->Cost)
					temp_rank++;
			}
		}
		population[i]->Rank = temp_rank;
    }

	fitnessSum = 0;
	for (i = 0; i < populationNumber; i++)
    {
         fitnessSum += population[i]->Rank;
    }



    while (!condition)
	{
		if (replicator == 0)
		{
			//roulette wheel
			if (threshold > 0)
				do{
					indv1  = rand()%(populationNumber-1);
					indv2  = rand()%(populationNumber-1);
				} while(population[indv1]->fitness > threshold &&  population[indv2]->fitness > threshold);
			else
				{
					first  = (1-(double(1.0*rand()/((double)RAND_MAX+1.0))))*fitnessSum;
					second  = (1-(double(1.0*rand()/((double)RAND_MAX+1.0))))*fitnessSum;
				}
		}
		else
		{
			//stochastic uniersal sampling
			if (threshold > 0)
				do{
					indv1  = rand()%(populationNumber-1);
					indv2  = rand()%(populationNumber-1);
				} while(indv1 > threshold &&  indv2 > threshold);
			else
				{
					first  = ((1-(double(1.0*rand()/((double)RAND_MAX+1.0))))*fitnessSum)/2;
					second = fitnessSum - first;
				}
		}

		if (threshold == 0)
		{
			tempsum = 0;
			for (int c = 0;c<populationNumber;c++)
			{
				tempsum += population[c]->Rank;
				if (tempsum >= first) 
				{
					indv1 = c; 
					break;
				}
			}
			
			tempsum = 0;
			for (int d = 0;d<populationNumber;d++)
			{
				tempsum += population[d]->Rank;
				if (tempsum >= second) 
				{
					indv2 = d; 
					break;
				}
			}
		}
		

		if(counter == populationNumber-1){
			  setIndv(new_population[counter], population[indv1]);
			counter++;
		} else
		{

			 setIndv(new_population[counter], population[indv1]);
			 setIndv(new_population[counter+1], population[indv2]);

			proba  = 1-(double(1.0*rand()/((double)RAND_MAX+1.0)));
			if (proba < proba_Crossover && indv1 != indv2)
				if (crossover == 1)
					apply1Crossover(counter, counter+1);
				else apply2Crossover(counter, counter+1);

			counter += 2;
		}
		if (counter >= populationNumber-1) condition = true;
	}
    for (i = 0;i<populationNumber;i++)
	   setIndv(population[i], new_population[i]);
}


void GA::setIndv(Individual *indv1, Individual *indv2)
{
		indv1->fitness = indv2->fitness;
		indv1->ioNumber = indv2->ioNumber;
		indv1->segmentNumber = indv2->segmentNumber;
		indv1->Cost = indv2->Cost;
		indv1->Error = indv2->Error;
		indv1->my_string = indv2->my_string;
}

void GA::reduceSegNumber(int a)
{
	int counter, n;

	counter = 0;
	n = 0;
	
	if (population[a]->ioNumber > 1)
	{
		while(n < population[a]->my_string.length())
		{
			if (population[a]->my_string.at(n) == 'p')
				counter++;
			if (counter >= 200)
			{
				population[a]->my_string.erase(n, population[a]->my_string.length());
				break;
			}
			n++;
		}
	}
}


void GA::makeFitness(Individual *I, qGate *finalG, double Error, double Cost)
{
	if (!grouped)
	{
		if (Error != 0)
		{
			//normalizing error
			I->Error = (Error / pow(2,(finalG->numIO*2)));
			
			//cost
			//setting real cost of minimized circuit
			I->Cost = Cost;

			//Baldwinian or not
			if (Ga != 0)
				Cost = minimizeCirc(I);

			//calculating cost of the circuit
			Cost = exp(-pow((divider-Cost),2));
			
			if (fitness == 0)
			{
				//simple fitness1
				I->fitness = 1-Error;
			} else 
				if (fitness == 1)
				{	
					//simple fitness1
					I->fitness = 1/(Error+1);
				}
				else
					if (fitness == 2)
					{
						//scaled complex fitness1
						I->fitness = alpha*(1-Error) + beta*Cost;
					}
					else
					{
						//scaled complex fitness2
						I->fitness = alpha*(1/(Error+1)) + beta*Cost;
					}
		}
		else 
		{
			I->Cost = Cost;
			I->Error = 0;
			I->fitness = 1;// - result/population[i].my_string.length();
		}
	} else
	{
		if (Error != 0)
		{
			//normalizing error
			I->Error = (Error / pow(2,(finalG->numIO*2)));

			//setting real cost of minimized circuit
			I->Cost = Cost;

			//Baldwinian or not
			if (Ga != 0)
				Cost = minimizeCirc(I);

			//calculating cost of the circuit
			Cost = exp(-pow((divider-Cost),2));


			if (fitness == 0)
			{
				//simple fitness1
				I->fitness = 1-Error;
			} else 
				if (fitness == 1)
				{	
					//simple fitness1
					I->fitness = 1/(Error+1);
				}
				else
					if (fitness == 2)
					{
						//scaled complex fitness1
						I->fitness = alpha*(1-Error) + beta*Cost;
					}
					else
					{
						//scaled complex fitness2
						I->fitness = alpha*(1/(Error+1)) + beta*Cost;
					}

			
			I->Groups++;
		}
		else 
		{
			//pow(2,finalG.numInput);
			I->Cost = Cost;
			I->Error = 0;
			I->fitness = 1;// - result/population[i].my_string.length();
		}
	}

}

bool GA::terminatingCondition()
{
     return condition;
}

void GA::setTerminatingCondition(int counter)
{
     if (counter >= MAXGEN)
	 {
		 out_stream<<" generation max reached "<<endl<<endl<<endl;
		 condition = true;
	 }
     else  if (getFitness()) 
			{
				out_stream<<" Solution found "<<endl<<endl<<endl;
				condition = true;
			}	
			else condition = false;
}



bool GA::getFitness()
{
	for (int a = 0;a<populationNumber;a++)
	{	
		if (population[a]->Error <= 0)
		{
			out_stream<<" best individual is "<<a<<endl<<endl;
			setIndv(&bestIndv, population[a]);
			return true;
		}
		else
			if (population[a]->fitness >= bestIndv.fitness)
			{
				setIndv(&bestIndv, population[a]);
			}
			else 
				if (population[a]->Error <= bestIndv.Error)
					setIndv(&bestIndv, population[a]);
	}
	return false;
}


void GA::initializeGates()
{
	I = dComp (0., 1.); //  Declare complex double variables
	double x, y;
	int counter = 0;
	string temp;
	
	//reading the input file 
	//first parameters
	in_stream>> populationNumber,"%\n";		//int
	in_stream >> proba_Mutation;		//double
	in_stream >> proba_Crossover;		//double
	in_stream >> alpha;					//float
	in_stream >> beta;					//float
	in_stream >> alpha1;				//float
	in_stream >> beta1;					//float
	in_stream >> divider;				//int
	//reading file divider
	in_stream >> temp;

	in_stream >> Ga;					//int
	in_stream >> mutation;				//int
	in_stream >> crossover;				//int
	in_stream >> replicator;			//int
	in_stream >> fitness;				//int
	in_stream >> grouped;				//bool
	in_stream >> pareto;				//bool
	in_stream >> threshold;				//double
	//reading file divider
	in_stream >> temp;

	out_stream<< " output file of synthesis  "<<endl;
	out_stream<< " size of the population					"<<populationNumber<<endl;
	out_stream<< " mutation probability is					"<< proba_Mutation<<endl;
	out_stream<< " mutation of crossover is					"<<proba_Crossover<<endl;
	out_stream<< " factor Alpha is							"<<alpha<<endl;
	out_stream<< " factor Beta is							"<<beta<<endl;
	out_stream<< " factor Alpha1 is							"<<alpha1<<endl;
	out_stream<< " factor Beta1 is							"<<beta1<<endl;
	out_stream<< " estimated minimum cost of this circuit	"<<divider<<endl;
	out_stream<< " -------------------------------------- "<<endl;
	out_stream<< " type of GA 0 - normal, 1 - Darwinian                      "<<Ga<<endl;
	out_stream<< " type of mutation 0 - normal, 1 - bitwise                  "<<mutation<<endl;
	out_stream<< " type of crossover 0 - 1point, 1 - 2point                  "<< crossover<<endl;
	out_stream<< " type of replication 0 - RW, 1 - SUS                       "<<replicator<<endl;
	out_stream<< " type of fitness 0 - simplest, 3 - complex                 "<<fitness<<endl;
	out_stream<< " type of fitness calculation 0 - individual, 1 - grouped   "<<grouped<<endl;
	out_stream<< " Pareto optimization 0 - no, 1 - yes                       "<< pareto<<endl;
	out_stream<< " Threshold replication 0 - no, other - threshold           "<<threshold<<endl;



	//bedining reading the gates
    in_stream >> numofgates;
	numofgates++;

	nullGate.numIO = 1;
	nullGate.representation = 'W';
	nullGate.gateMatrix1[0][0] = 1;
	nullGate.gateMatrix1[0][1] = 0;
	nullGate.gateMatrix1[1][0] = 0;
	nullGate.gateMatrix1[1][1] = 1;
	nullGate.Cost = 0;
	gateArray[0] = &nullGate;

	out_stream<< gateArray[0]->representation <<endl;
	out_stream<<"  Cost of the gate  "<< gateArray[0]->Cost <<endl;
	out_stream<<"  IO of the gate  "<< gateArray[0]->numIO <<endl;

	for (int j = 0; j < (int(pow(2,gateArray[0]->numIO))); j++)
	{
		for (int k = 0; k < (int(pow(2,gateArray[0]->numIO))); k++)
		{
			out_stream<< gateArray[0]->gateMatrix1[j][k]<<' ';
		}
		out_stream<< endl;
	}
	
	// get first gate
    for (int a = 1; a < numofgates; a++)
	{
		gateArray[a] = new qGate;
		//cout<<gateArray[a]<<endl;
		
		in_stream >> gateArray[a]->numIO;
		gateArray[a]->representation = char(a+65); // starting from A
		in_stream >> gateArray[a]->Cost;
		
		out_stream<< gateArray[a]->representation <<endl;
		out_stream<<"  Cost of the gate  "<< gateArray[a]->Cost <<endl;
		out_stream<<"  IO of the gate  "<< gateArray[a]->numIO <<endl;
		
		for (int j = 0; j < (int(pow(2,gateArray[a]->numIO))); j++)
		{
			for (int k = 0; k < (int(pow(2,gateArray[a]->numIO))); k++)
			{
				in_stream >> x;
				out_stream<< x <<' ';
				if (x == 2)
					gateArray[a]->gateMatrix1[j][k] = I;
				else if (x == -2)
					gateArray[a]->gateMatrix1[j][k] = -I;
				else if (x == 4)
					gateArray[a]->gateMatrix1[j][k] = exp(complex<double>(0, (3.1416/4)));
				else if (x == 5)
					gateArray[a]->gateMatrix1[j][k] = complex<double>(0.5, 0.5);
				else if (x == -5)
					gateArray[a]->gateMatrix1[j][k] = complex<double>(0.5, -0.5);
				else gateArray[a]->gateMatrix1[j][k] = x;
				out_stream<< gateArray[a]->gateMatrix1[j][k]<<' ';
			}
			out_stream<< endl;
		}
		counter++;
	}
	
	// get the expected result matrix from the file
	in_stream >> resultnum;
	finalGate.numIO = resultnum;
	
	out_stream<< "final gate "<<endl;
	
	for (int m = 0; m < (int(pow(2,finalGate.numIO))); m++)
	{
		for (int n = 0; n < (int(pow(2,finalGate.numIO))); n++)
		{
			in_stream >> y;
			if (y == 2)
				finalGate.gateMatrix1[m][n] = I;
			else if (y == -2)
				finalGate.gateMatrix1[m][n] = -I;
			else finalGate.gateMatrix1[m][n] = y;
			out_stream<< finalGate.gateMatrix1[m][n];
		}
		out_stream<<endl;
	}
	out_stream<< "------------------------------------------"<<endl;  
}


qGate GA::tensorProduct(qGate *A, qGate *B)
{
    qGate pGate;
	for (int i = 0; i < (int(pow(2,A->numIO))); i++)
	{    
		for (int j = 0; j < (int(pow(2,A->numIO))); j++)
		{   
			for (int k = 0; k < (int(pow(2,B->numIO))); k++)
			{   
				for (int l = 0; l < (int(pow(2,B->numIO))); l++)
				{  
					pGate.gateMatrix1[i*(int(pow(2,B->numIO)))+k][j*(int(pow(2,B->numIO)))+l] = A->gateMatrix1[i][j] * B->gateMatrix1[k][l];
				}
			}
		}
	}
	pGate.numIO = (A->numIO + B->numIO);
	return pGate;
}


qGate GA::matrixProduct(qGate *A, qGate *B)
{
	qGate pGate;
	//rows of the first gate
	for (int i = (int(pow(2,A->numIO))-1);i>=0; i--)
	{   //colums of the second gate
		for (int j = 0; j < (int(pow(2,B->numIO))); j++)
		{   //colums of the first matrix
			int l = (int(pow(2,A->numIO))); 
			for (int k = 0; k < (int(pow(2,A->numIO))); k++)
			{ //rows of the second
				pGate.gateMatrix1[j][i] = pGate.gateMatrix1[j][i] + A->gateMatrix1[k][i] * B->gateMatrix1[j][k];
				l--;
			}
		}
	}
	pGate.numIO = A->numIO;
	return pGate;
}

void GA::closeStream()
{
     in_stream.close();
     out_stream.close();
}

////////////////////////////////////////////////////
////////////////////////////////////////////////////
////////////////////////////////////////////////////
//methods of the minimizer -- yet all in one class
//no time for improvements
////////////////////////////////////////////////////
////////////////////////////////////////////////////
////////////////////////////////////////////////////
qGate GA::copyGate(qGate *Circ)
{
		qGate Circf;

		Circf = *Circ;
		Circf.numIO = Circ->numIO;
		for (int i = 0; i < (int(pow(2,Circ->numIO))); i++)
		    for (int j = 0; j < (int(pow(2,Circ->numIO))); j++)
                      Circf.gateMatrix1[i][j] = Circ->gateMatrix1[i][j];

		Circf.Cost = Circ->Cost;
		Circf.representation= Circ->representation;

		return Circf;
}

int GA::compareGate(qGate *A)
{
	int a = 0;
	bool matrixTest;
	int result = -1;
	while (a<numofgates)
	{
		out_stream<< a<<" "<<numofgates<<endl;
		matrixTest = true;

		if (A->numIO != gateArray[a]->numIO)
			result = -1;
		else
		{
			for (int i = 0; i < (int(pow(2,A->numIO))); i++)
				for (int j = 0; j < (int(pow(2,A->numIO))); j++)
                      if (A->gateMatrix1[i][j] != gateArray[a]->gateMatrix1[i][j])
						  matrixTest = false;
			if (matrixTest) 
				result = a;
			else result = -1;

		}

		if (result != -1) break;
		else a++;
	}

	return result;
	
}

bool GA::compareGate(qGate A, qGate B)
{
	if (A.numIO != B.numIO)
		return false;
	else
	{
		for (int i = 0; i < (int(pow(2,A.numIO))); i++)
			for (int j = 0; j < (int(pow(2,B.numIO))); j++)
				if (A.gateMatrix1[i][j] != B.gateMatrix1[i][j])
					return false;
	}
	return true;
	
}
// the super-method intializing the circuit minimization
// now the minimzation is done here only virtually
int GA::minimizeCirc(Individual *qC)
{
	int begin0, begin1, end0, end1;
	bool equal;
	int a, b;
	int result = 0;

	if (qC->ioNumber > 1)
	{	
		begin0 = 0;
		end0 = qC->my_string.find("p", begin0+1)+1;
		begin1 = end0;
		end1 = qC->my_string.find("p", begin1+1)+1;
		b = 0;

		while (end1 > 0)
		{
			equal = false;
			a = 0;
			if ((qC->my_string.substr(begin0, (end0-begin0))).length() == (qC->my_string.substr(begin1, (end1-begin1))).length() )
			{
				while(a <= (end0-begin0-2))
				{
					if ( (qC->my_string.at(begin0+a)) != 'p')
					{	
						if (gateArray[getGate(qC->my_string.at(begin0+a))]->numIO != gateArray[getGate(qC->my_string.at(begin1+a))]->numIO)
						{
							equal = false;
							break;
						}
						else 
						{
							if (a >= end0-begin0-2) 
							{
								equal = true;
							}
						}
					}
					a++;
				}
			}
			if (!equal)
			{
				for (int a = 1;a<(end0-begin0-1);a++)
				{
					result += gateArray[getGate(qC->my_string.at(begin0+a))]->Cost;
				}
			}
			begin0 = begin1;
			end0 = qC->my_string.find("p", begin0+1)+1;
			begin1 = end0;
			end1 = qC->my_string.find("p", begin1+1)+1;
		}
	}
	for (a = 1;a<(end0-begin0-1);a++)
	{
		result += gateArray[getGate(qC->my_string.at(begin0+a))]->Cost;
	}
	qC->Cost = result;

	return result;
}

//removes physically redundant segments in the concenred circuit
void GA::removeRedund(Individual *C)
{
	int begin0, begin1, end0, end1;
	bool equal, erased;
	int a, b;

	if (C->ioNumber > 1)
	{	
		begin0 = 0;
		end0 = C->my_string.find("p", begin0+1)+1;
		begin1 = end0;
		end1 = C->my_string.find("p", begin1+1)+1;
		b = 0;

		while (end0<C->my_string.length()-3 || end0 < 0)
		{
			equal = false;
			erased = false;
			a = 0;
			if (C->my_string.substr(begin0, (end0-begin0)) == C->my_string.substr(begin1, (end1-begin1)))
			{
				C->my_string.erase(begin0, (end1-begin0));
				erased = true;
			}
	

			if (erased)
			{
				if (begin0 != 0)
				{
					begin0 = 0;
					end0 = C->my_string.find("p", begin0+1)+1;
					
					begin1 = end0;
					end1 = C->my_string.find("p", begin1+1)+1;
				}
				else 
				{
					begin1 = end0;
					end1 = C->my_string.find("p", begin1+1)+1;
					erased = false;
				}
			}
			else 
			{
				begin0 = begin1;
				end0 = C->my_string.find("p", begin0+1)+1;
				if (end0 < C->my_string.length()-3)
				{
					begin1 = end0;
					end1 = C->my_string.find("p", begin1+1)+1;
				}
			}
		}	
	}
}


//physically modyfies the circuit for lamarckian learning
void GA::mergeSegs(Individual *Circ, int first, int second)
{
	basic_string<char> tempStr;
	qGate *temp, *fGate, *sGate;
	int gateCount;

	int endfrsSg = Circ->my_string.find("p", first+1)+1;
	int endscdSg = Circ->my_string.find("p", second+1)+1;
	tempStr.append(1, 'p');

	for (int a = 0;a<(endfrsSg-first-1);a++)
	{
		if (Circ->my_string.at(first+a) != 'p')
		{
			fGate = gateArray[getGate(Circ->my_string.at(first+a))];
			sGate = gateArray[getGate(Circ->my_string.at(second+a))];
			temp = &matrixProduct(fGate, sGate);
			gateCount = compareGate(temp);

			if (gateCount != -1)
			{
				tempStr.append(1, gateArray[gateCount]->representation);
			}
			else
			{
				gateArray[numofgates] = new qGate;
				(*gateArray[numofgates]) = copyGate(temp);

				if (gateArray[numofgates-1]->representation == 'o')
					gateArray[numofgates]->representation = gateArray[numofgates-1]->representation+char(2);
				else if (gateArray[numofgates-1]->representation == char(264))
					gateArray[numofgates]->representation = gateArray[numofgates-1]->representation+char(6);
					else gateArray[numofgates]->representation = gateArray[numofgates-1]->representation+char(1);

				gateArray[numofgates]->Cost = fGate->Cost;
				gateArray[numofgates]->parentA = fGate->representation;
				gateArray[numofgates]->parentB = fGate->representation;
				tempStr.append(1, gateArray[numofgates]->representation);
				numofgates++;
			}
		}
	}
	tempStr.append(1, 'p');
	Circ->my_string.erase(first, (endscdSg-first));
	Circ->my_string.insert(first,tempStr);
}


